@@ -2,11 +2,10 @@ package PHP::Functions::Password;
 use strict;
 use warnings;
 use Carp qw(carp croak);
-use Crypt::Eksblowfish ();
+use Crypt::Bcrypt ();
 use Crypt::OpenSSL::Random ();
-use MIME::Base64 qw(encode_base64 decode_base64);
+use MIME::Base64 qw(decode_base64);
 use Readonly qw(Readonly);
-use version 0.77 ();
 use base qw(Exporter);
 
 our @EXPORT;
@@ -22,69 +21,69 @@ our @EXPORT_OK = qw(
 	PASSWORD_DEFAULT
 );
 our %EXPORT_TAGS = (
-	'all'		=> \@EXPORT_OK,
-	'default'	=> \@EXPORT,
-	'consts'	=> [ grep /^PASSWORD_/, @EXPORT_OK ],
-	'funcs'		=> [ grep /^password_/, @EXPORT_OK ],
+	'all'     => \@EXPORT_OK,
+	'default' => \@EXPORT,
+	'consts'  => [ grep /^PASSWORD_/, @EXPORT_OK ],
+	'funcs'   => [ grep /^password_/, @EXPORT_OK ],
 );
-our $VERSION = '1.12';
+our $VERSION = '1.13';
 
 
 # Exported constants
 use constant PASSWORD_BCRYPT   => 1;
-use constant PASSWORD_ARGON2I  => 2;
-use constant PASSWORD_ARGON2ID => 3;
+use constant PASSWORD_ARGON2I  => 2; # exists in PHP since version 7.2
+use constant PASSWORD_ARGON2ID => 3; # exists in PHP since version 7.3
 use constant PASSWORD_DEFAULT  => PASSWORD_BCRYPT;
 
 
 # Internal constants
-Readonly my $PASSWORD_BCRYPT_DEFAULT_COST => 10;	# no such PHP constant
-Readonly my $PASSWORD_ARGON2_DEFAULT_SALT_LENGTH => 16;	# no such PHP constant
+Readonly my $PASSWORD_BCRYPT_DEFAULT_COST => 10;        # no such PHP constant
+Readonly my $PASSWORD_BCRYPT_MAX_PASSWORD_LEN => 72;    # no such PHP constant
+Readonly my $PASSWORD_ARGON2_DEFAULT_SALT_LENGTH => 16; # no such PHP constant
 Readonly my $PASSWORD_ARGON2_DEFAULT_MEMORY_COST => 65536;
 Readonly my $PASSWORD_ARGON2_DEFAULT_TIME_COST => 4;
 Readonly my $PASSWORD_ARGON2_DEFAULT_THREADS => 1;
-Readonly my $PASSWORD_ARGON2_DEFAULT_TAG_LENGTH => 32;	# no such PHP constant
+Readonly my $PASSWORD_ARGON2_DEFAULT_TAG_LENGTH => 32;  # no such PHP constant
 
-Readonly my $SIG_BCRYPT => '2y';	# PHP default
-Readonly my $SIG_ARGON2I  => 'argon2i';
-Readonly my $SIG_ARGON2ID => 'argon2id';
+Readonly my $SIG_BCRYPT    => '2y'; # PHP default; equivalent of 2b in non-PHP implementations
+Readonly my $SIG_ARGON2I   => 'argon2i';
+Readonly my $SIG_ARGON2ID  => 'argon2id';
 
-Readonly my %SIG_TO_ALGO => (
-	$SIG_BCRYPT		=> PASSWORD_BCRYPT,
-	$SIG_ARGON2I	=> PASSWORD_ARGON2I,
-	$SIG_ARGON2ID	=> PASSWORD_ARGON2ID,
+Readonly my %SIG_TO_ALGO => (	# not used for bcrypt
+	$SIG_ARGON2I   => PASSWORD_ARGON2I,
+	$SIG_ARGON2ID  => PASSWORD_ARGON2ID,
 );
 
 # https://en.wikipedia.org/wiki/Bcrypt
 Readonly my $RE_BCRYPT_ALGO => qr#2[abxy]?#;
-Readonly my $RE_BCRYPT_SALT => qr#[./A-Za-z0-9]{22}#;	# fixed 16 byte salt
+Readonly my $RE_BCRYPT_SALT => qr#[./A-Za-z0-9]{22}#;	# fixed 16 byte salt (encoded as 22 bcrypt-custom-base64 chars)
 Readonly my $RE_BCRYPT_COST => qr#[0-3]\d#;
 Readonly my $RE_BCRYPT_HASH => qr#[./A-Za-z0-9]+#;
 Readonly my $RE_BCRYPT_STRING => qr/^
 	\$
-	($RE_BCRYPT_ALGO)	# $1 type
+	($RE_BCRYPT_ALGO)  # $1 type
 	\$
-	($RE_BCRYPT_COST)	# $2 cost
+	($RE_BCRYPT_COST)  # $2 cost
 	\$
-	($RE_BCRYPT_SALT)	# $3 salt
-	($RE_BCRYPT_HASH)	# $4 hash
+	($RE_BCRYPT_SALT)  # $3 salt
+	($RE_BCRYPT_HASH)  # $4 hash
 $/x;
 
 # See https://www.alexedwards.net/blog/how-to-hash-and-verify-passwords-with-argon2-in-go
 Readonly my $RE_ARGON2_ALGO => qr#argon2id?#;
 Readonly my $RE_ARGON2_STRING => qr/^
 	\$
-	($RE_ARGON2_ALGO)	# $1 signature
+	($RE_ARGON2_ALGO)   # $1 signature
 	\$
-	v=(\d{1,3})			# $2 version
+	v=(\d{1,3})         # $2 version
 	\$
-		m=(\d{1,10}),	# $3 memory_cost
-		t=(\d{1,3}),	# $4 time_cost
-		p=(\d{1,3})		# $5 threads
+		m=(\d{1,10}),   # $3 memory_cost
+		t=(\d{1,3}),    # $4 time_cost
+		p=(\d{1,3})     # $5 threads
 	\$
-	([A-Za-z0-9+\/]+)	# $6 salt
+	([A-Za-z0-9+\/]+)   # $6 salt
 	\$
-	([A-Za-z0-9+\/]+)	# $7 hash
+	([A-Za-z0-9+\/]+)   # $7 hash
 $/x;
 
 =head1 NAME
@@ -106,7 +105,7 @@ Functional interface, typical use:
 
 	use PHP::Functions::Password qw(password_hash);
 	my $password = 'secret';
-	my $crypted_string = password_hash($password);	# uses PASSWORD_BCRYPT algorithm
+	my $crypted_string = password_hash($password);  # uses PASSWORD_BCRYPT algorithm
 
 Functional interface use, using options:
 
@@ -119,6 +118,13 @@ Functional interface use, using options:
 	# Use a different algorithm:
 	my $crypted_string = password_hash($password, PASSWORD_ARGON2ID);
 
+	# Better practice using a 'pepper':
+	use Digest::SHA qw(hmac_sha256);
+	my $pepper = 'Abracadabra and Hocus pocus';  # retrieve this from a secrets config file for example (and don't loose it!)
+	my $peppered_password = hmac_sha256($password, $pepper);
+	my $crypted_string = password_hash($password, PASSWORD_ARGON2ID);  # store this in your database
+	# ... and when verifying passwords, then you pepper then first too.
+
 Class method use, using options:
 
 	use PHP::Functions::Password;
@@ -140,9 +146,9 @@ The following names can be imported into the calling namespace by request:
 	PASSWORD_ARGON2ID
 	PASSWORD_BCRYPT
 	PASSWORD_DEFAULT
-	:all	- what it says
-	:consts	- the PASSWORD_* constants
-	:funcs	- the password_* functions
+	:all    - what it says
+	:consts - the PASSWORD_* constants
+	:funcs  - the password_* functions
 
 =head1 PHP COMPATIBLE AND EXPORTABLE FUNCTIONS
 
@@ -190,14 +196,14 @@ sub password_get_info {
 		my $salt = $3;
 		my $hash = $4;
 		my %result = (
-			'algo'		=> PASSWORD_BCRYPT,
-			'algoName'	=> 'bcrypt',
-			'options'	=> {
-				'cost'	=> $cost,
+			'algo'     => PASSWORD_BCRYPT,
+			'algoName' => 'bcrypt',
+			'algoSig'  => $type,  # extra
+			'options'  => {
+				'cost'=> $cost,
 			},
-			'algoSig'	=> $type,	# extra
-			'salt'		=> $salt,	# extra
-			'hash'		=> $hash,	# extra
+			'salt' => $salt,  # extra
+			'hash' => $hash,  # extra
 		);
 		return wantarray ? %result : \%result;
 	}
@@ -212,26 +218,26 @@ sub password_get_info {
 		#my $raw_salt = decode_base64($salt);
 		#my $raw_hash = decode_base64($hash);
 		my %result = (
-			'algo'		=> $SIG_TO_ALGO{$sig},
-			'algoName'	=> $sig,
-			'options'	=> {
-				'memory_cost'	=> $memory_cost,
-				'time_cost'		=> $time_cost,
-				'threads'		=> $threads,
+			'algo'     => $SIG_TO_ALGO{$sig},
+			'algoName' => $sig,
+			'algoSig'  => $sig,
+			'options'  => {
+				'memory_cost' => $memory_cost,
+				'time_cost'   => $time_cost,
+				'threads'     => $threads,
 			},
-			'algoSig'	=> $sig,
-			'salt'		=> $salt,
-			'hash'		=> $hash,
-			'version'	=> $version,
+			'salt'    => $salt,
+			'hash'    => $hash,
+			'version' => $version,
 		);
 		return wantarray ? %result : \%result;
 	}
 
 	# No matches:
 	my %result = (
-		'algo'		=> 0,
-		'algoName'	=> 'unknown',
-		'options'	=> {},
+		'algo'     => 0,
+		'algoName' => 'unknown',
+		'options'  => {},
 	);
 	return wantarray ? %result : \%result;
 }
@@ -241,12 +247,11 @@ sub password_get_info {
 
 =item password_hash($password, $algo, %options)
 
-Similar to L<http://php.net/manual/en/function.password-hash.php>
-with difference that the $algo argument is optional and defaults to PASSWORD_DEFAULT for your programming pleasure.
+Similar to L<http://php.net/manual/en/function.password-hash.php> with the difference that the $algo argument is optional and defaults to PASSWORD_DEFAULT for your programming pleasure.
 
-Important notes about the 'salt' option which you shouldn't use:
+Important notes about the 'salt' option which you shouldn't use in the first place:
 
-	- The PASSWORD_BCRYPT 'salt' option is deprecated since PHP 7.0, but if you do pass it, then it must be bcrypt custom base64 encoded and not raw bytes!
+	- The PASSWORD_BCRYPT 'salt' option is deprecated since PHP 7.0, but if you do pass it, then it must be 16 bytes long!
 	- For algorithms other than PASSWORD_BCRYPT, PHP doesn't support the 'salt' option, but if you do pass it, then it must be in raw bytes!
 
 =cut
@@ -256,20 +261,50 @@ sub password_hash {
 	my $password = shift;
 	my $algo = shift // PASSWORD_DEFAULT;
 	my %options = @_ && ref($_[0]) ? %{$_[0]} : @_;
+	unless ($algo =~ /^\d$/) {
+		croak("Invalid \$algo parameter ($algo) which should be one of the PASSWORD_* integer constants");
+	}
 	if ($algo == PASSWORD_BCRYPT) {
 		my $salt;
-		if (defined($options{'salt'}) && length($options{'salt'})) {	# bcrypt custom base64 encoded!
-			unless ($options{'salt'} =~ /^$RE_BCRYPT_SALT$/) {
+		if (defined($options{'salt'}) && length($options{'salt'})) {
+			# Treat salt as a string of bytes
+			$salt = $options{'salt'};
+			utf8::is_utf8($salt) && utf8::encode($salt);	# "\x{100}"  becomes "\xc4\x80"; preferred equivalent of Encode::is_utf8($string) && Encode::_utf8_off($password);
+			if (length($salt) == 16) {
+				# raw bytes: OK
+			}
+			elsif ($salt =~ /^$RE_BCRYPT_SALT$/) {	# bcrypt-custom-base64 encoded string of 22 characters; DEPRECATED
+				$salt =~ tr#./A-Za-z0-9#A-Za-z0-9+/#;
+				$salt .= '=' x (3 - (length($salt) + 3) % 4);
+				$salt = decode_base64($salt);
+			}
+			else {
 				croak('Bad syntax in given and deprecated salt option (' . $options{'salt'} . ')');
 			}
-			$salt = $options{'salt'};
 		}
 		else {
-			$salt = $proto->_bcrypt_base64_encode(Crypt::OpenSSL::Random::random_bytes(16));
+			$salt = Crypt::OpenSSL::Random::random_bytes(16);
 		}
-		my $cost = $options{'cost'} || $PASSWORD_BCRYPT_DEFAULT_COST;
-		my $settings = '$' . $SIG_BCRYPT . '$' . sprintf('%.2u', $cost) . '$' . $salt;
-		return $proto->_bcrypt($password, $settings);
+		my $cost = $PASSWORD_BCRYPT_DEFAULT_COST;
+		if ($options{'cost'}) {
+			my $min_cost = 5;
+			my $max_cost = 31;
+			unless (($options{'cost'} =~ /^\d{1,2}$/) && ($options{'cost'} >= $min_cost) && ($options{'cost'} <= $max_cost)) {
+				croak('Invalid cost option given (' . $options{'cost'} . ") which should be an integer in the range $min_cost to $max_cost");
+			}
+			$cost = int($options{'cost'});
+		}
+
+		# Treat passwords as strings of bytes
+		utf8::is_utf8($password) && utf8::encode($password);	# "\x{100}"  becomes "\xc4\x80"; preferred equivalent of Encode::is_utf8($string) && Encode::_utf8_off($password);
+
+		# Everything beyond the max password length in bytes for bcrypt is silently ignored.
+		require bytes;
+		if (bytes::length($password) > $PASSWORD_BCRYPT_MAX_PASSWORD_LEN) {	# $password is already bytes, so the bytes:: prefix is redundant here
+			$password = substr($password, 0, $PASSWORD_BCRYPT_MAX_PASSWORD_LEN);
+		}
+
+		return Crypt::Bcrypt::bcrypt($password, $SIG_BCRYPT, $cost, $salt);
 	}
 	elsif (($algo == PASSWORD_ARGON2ID) || ($algo == PASSWORD_ARGON2I)) {
 		unless ($INC{'Crypt/Argon2.pm'} || eval { require Crypt::Argon2; }) {
@@ -282,7 +317,7 @@ sub password_hash {
 		my $threads = $options{'threads'} || $PASSWORD_ARGON2_DEFAULT_THREADS;
 		my $tag_length = $options{'tag_length'} || $PASSWORD_ARGON2_DEFAULT_TAG_LENGTH;	# undocumented; not a PHP option; 4 - 2^32 - 1
 
-		# Ignore characters and treat strings as strings of bytes
+		# Treat passwords as strings of bytes
 		utf8::is_utf8($password) && utf8::encode($password);	# "\x{100}"  becomes "\xc4\x80"; preferred equivalent of Encode::is_utf8($string) && Encode::_utf8_off($password);
 
 		my @args = ($password, $salt, $time_cost, $memory_cost . 'k', $threads, $tag_length);
@@ -318,7 +353,8 @@ sub password_needs_rehash {
 		return 1;
 	}
 	if ($algo == PASSWORD_BCRYPT) {
-		if ($info{'algoSig'} ne $SIG_BCRYPT) {
+		#unless (($info{'algoSig'} eq $SIG_BCRYPT) || ($info{'algoSig'} eq '2b')) {	# also accept 2b as a non-PHP equivalent of 2y
+		unless ($info{'algoSig'} eq $SIG_BCRYPT) {	# this emulates PHP's behaviour (it requires 2b to be rehashed as 2y).
 			$options{'debug'} && warn('Algorithm signatures differ: ' . $info{'algoSig'} . ' vs ' . $SIG_BCRYPT);
 			return 1;
 		}
@@ -348,7 +384,10 @@ sub password_needs_rehash {
 		my $wanted_tag_length = $options{'tag_length'} || $PASSWORD_ARGON2_DEFAULT_TAG_LENGTH;	# undocumented; not a PHP option; 4 - 2^32 - 1
 
 		if ($INC{'Crypt/Argon2.pm'} || eval { require Crypt::Argon2; }) {
-			if (version->parse($Crypt::Argon2::VERSION) < version->parse('0.008')) {
+			if (Crypt::Argon2->can('argon2_needs_rehash')) {	# since version 0.008
+				return Crypt::Argon2::argon2_needs_rehash($crypted, $info{'algoSig'}, $time_cost, $memory_cost . 'k', $threads, $wanted_tag_length, $wanted_salt_length);
+			}
+			else {	# as long as Crypt::Argon2 is not required for building, a minimum version requirement cannot be forced, and therefore the workaround below is needed
 				if ($info{'version'} < 19) {
 					$options{'debug'} && warn('Version mismatch: ' . $info{'version'} . '<19');
 					return 1;
@@ -372,9 +411,6 @@ sub password_needs_rehash {
 					return 1;
 				}
 			}
-			else {
-				return Crypt::Argon2::argon2_needs_rehash($crypted, $info{'algoSig'}, $time_cost, $memory_cost . 'k', $threads, $wanted_tag_length, $wanted_salt_length);
-			}
 		}
 	}
 	else {
@@ -396,19 +432,17 @@ sub password_verify {
 	my $proto = @_ && UNIVERSAL::isa($_[0],__PACKAGE__) ? shift : __PACKAGE__;
 	my ($password, $crypted) = @_;
 	if ($crypted =~ $RE_BCRYPT_STRING) {
-		my $cost = int($2);
-		my $salt = $3;
-		my $hash = $4;
-		my $new_crypt = $proto->hash(
-			$password,
-			'cost'	=> $cost,
-			'salt'	=> $salt,
-		);
-		if ($crypted eq $new_crypt) {
-			return 1;
+
+		# Treat passwords as strings of bytes
+		utf8::is_utf8($password) && utf8::encode($password);	# "\x{100}"  becomes "\xc4\x80"; preferred equivalent of Encode::is_utf8($string) && Encode::_utf8_off($password);
+
+		# Everything beyond the max password length in bytes for bcrypt is silently ignored.
+		require bytes;
+		if (bytes::length($password) > $PASSWORD_BCRYPT_MAX_PASSWORD_LEN) {	# $password is already bytes, so the bytes:: prefix is redundant here
+			$password = substr($password, 0, $PASSWORD_BCRYPT_MAX_PASSWORD_LEN);
 		}
-		# Since the signature may vary slightly, try comparing only the hash.
-		return ($new_crypt =~ $RE_BCRYPT_STRING) && ($4 eq $hash);
+
+		return Crypt::Bcrypt::bcrypt_check($password, $crypted);
 	}
 	elsif ($crypted =~ $RE_ARGON2_STRING) {
 		unless ($INC{'Crypt/Argon2.pm'} || eval { require Crypt::Argon2; }) {
@@ -417,7 +451,7 @@ sub password_verify {
 		}
 		my $algo = $SIG_TO_ALGO{$1};
 
-		# Ignore characters and treat strings as strings of bytes
+		# Treat passwords as strings of bytes
 		utf8::is_utf8($password) && utf8::encode($password);	# "\x{100}"  becomes "\xc4\x80"; preferred equivalent of Encode::is_utf8($string) && Encode::_utf8_off($password);
 
 		my @args = ($crypted, $password);
@@ -521,106 +555,13 @@ sub verify {
 
 
 
-
-# From Crypt::Eksblowfish::Bcrypt.
-# This is a version of C<crypt> (see L<perlfunc/crypt>) that implements the bcrypt algorithm.
-sub _bcrypt {
-	my $proto = @_ && UNIVERSAL::isa($_[0],__PACKAGE__) ? shift : __PACKAGE__;
-	my ($password, $settings) = @_;
-	unless ($settings =~ qr/^
-		\$
-		($RE_BCRYPT_ALGO)
-		\$
-		($RE_BCRYPT_COST)
-		\$
-		($RE_BCRYPT_SALT)
-	/x) {
-		croak('Bad bcrypt settings argument');
-	}
-	my ($type, $cost, $salt_base64) = ($1, $2, $3);
-	my $hash = $proto->_bcrypt_hash(
-		$password,
-		{
-			'key_nul'	=> length($type) > 1,
-			'cost'		=> $cost,
-			'salt'		=> $proto->_bcrypt_base64_decode($salt_base64),
-		}
-	);
-	return '$' . $SIG_BCRYPT . '$' . $cost . '$' . $salt_base64 . $proto->_bcrypt_base64_encode($hash);
-}
-
-
-
-
-# From Crypt::Eksblowfish::Bcrypt.
-# Hashes $password according to the supplied $settings, and returns the 23-octet hash.
-sub _bcrypt_hash {
-	my $proto = @_ && UNIVERSAL::isa($_[0],__PACKAGE__) ? shift : __PACKAGE__;
-	my ($password, $settings) = @_;
-	if ($settings->{'key_nul'} || ($password eq '')) {
-		$password .= "\0";
-	}
-
-	# Ignore characters and treat strings as strings of bytes
-	utf8::is_utf8($password) && utf8::encode($password);	# "\x{100}"  becomes "\xc4\x80"; preferred equivalent of Encode::is_utf8($string) && Encode::_utf8_off($password);
-
-	my $cipher = Crypt::Eksblowfish->new(
-		$settings->{'cost'},
-		$settings->{'salt'},
-		substr($password, 0, 72)
-	);
-	my $hash = join('',
-		map {
-			my $blk = $_;
-			for(my $i = 64; $i--; ) {
-				$blk = $cipher->encrypt($blk);
-			}
-			$blk;
-		} qw(OrpheanB eholderS cryDoubt)
-	);
-	chop($hash);
-	return $hash;
-}
-
-
-
-
-# From Crypt::Eksblowfish::Bcrypt.
-# Decodes an octet string that was textually encoded using the form of base64 that is conventionally used with bcrypt.
-sub _bcrypt_base64_decode {
-	my $proto = @_ && UNIVERSAL::isa($_[0],__PACKAGE__) ? shift : __PACKAGE__;
-	my $text = shift;
-	unless ($text =~ m!\A(?>(?:[./A-Za-z0-9]{4})*)(?:|[./A-Za-z0-9]{2}[.CGKOSWaeimquy26]|[./A-Za-z0-9][.Oeu])\z!) {
-		croak('Bad base64 encoded text argument');
-	}
-	$text =~ tr#./A-Za-z0-9#A-Za-z0-9+/#;
-	$text .= '=' x (3 - (length($text) + 3) % 4);
-	return decode_base64($text);
-}
-
-
-
-
-# From Crypt::Eksblowfish::Bcrypt.
-# Encodes the octet string textually using the form of base64 that is conventionally used with bcrypt.
-sub _bcrypt_base64_encode {
-	my $proto = @_ && UNIVERSAL::isa($_[0],__PACKAGE__) ? shift : __PACKAGE__;
-	my $octets = shift;
-	my $text = encode_base64($octets, '');
-	$text =~ tr#A-Za-z0-9+/=#./A-Za-z0-9#d;	# "=" padding is deleted
-	return $text;
-}
-
-
-
 1;
 
 __END__
 
 =head1 SEE ALSO
 
- L<Crypt::Eksblowfish::Bcrypt> from which several internal functions were copied and slightly modified.
- L<Crypt::Eksblowfish> used for creating/verifying crypted strings in bcrypt format.
+ L<Crypt::Bcrypt> used for all the bcrypt support.
  L<Crypt::OpenSSL::Random> used for random salt generation.
  L<Crypt::Argon2> recommended for argon2 algorithm support.
 
